/*
 *		Amiga Linux/68k Loader
 *
 *		 Copyright 1995 by Geert Uytterhoeven (Geert.Uytterhoeven@cs.kuleuven.ac.be)
 *
 *		-----------------------------------------------------------------------------
 *
 *		Configuration file parsing based on
 *
 *			alilo.c - Auto Folder LILO emulation for Atari
 *
 *			Copyright 1994 Roman Hodek
 *
 *
 *		The original idea for this program comes from the Linux distribution:
 *
 *			LILO  -  Generic Boot Loader for Linux ("LInux LOader")
 *						by Werner Almesberger
 *
 *		-----------------------------------------------------------------------------
 *
 *		This file is subject to the terms and conditions of the GNU General Public
 *		License.  See the file README.legal in the main directory of the Linux/68k
 *		distribution for more details.
 */


#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <dirent.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <linux/fs.h>
#include <linux/amigatypes.h>
#include <linux/amihdreg.h>
#include <linux/major.h>
#include <linux/bootinfo.h>

#include "lilo.h"


#define VERSION "Amiga Linux/68k Loader Version 0.2 (14/02/95)"


	/*
	 *		Command Line Options
	 */

char *Name = NULL;
int Verbose = 0;
int Install = 1;
int Force = 0;
char *DosType = NULL;
char *SaveBootBlock = NULL;
char *RestoreBootBlock = NULL;
char *Device = NULL;


	/*
	 *		Boot Code Definitions
	 */

extern char BootCodeStart[];
extern char BootCodeEnd[];
extern char LoaderVectorStart[];
extern char LoaderVectorEnd[];


	/*
	 *		Valid DosTypes
	 */

struct DosTypeEntry {
	u_long ID;
	char *ASCII;
	char *Description;
};

struct DosTypeEntry DosTypeEntries[] = {
	{ ID_BOOT_DISK, "BOOU", "Generic boot disk" },
	{ ID_OFS_DISK, "DOS0", "OFS disk" },
	{ ID_FFS_DISK, "DOS1", "FFS disk" },
	{ ID_OFS_INTL_DISK, "DOS2", "OFS INTL disk" },
	{ ID_FFS_INTL_DISK, "DOS3", "FFS INTL disk" },
	{ ID_OFS_DC_DISK, "DOS4", "OFS DC disk" },
	{ ID_FFS_DC_DISK, "DOS5", "FFS DC disk" },
	{ ID_MUFS_DISK, "muFS", "MultiUser FFS INTL disk" },
	{ ID_MU_OFS_DISK, "muF0", "MultiUser OFS disk" },
	{ ID_MU_FFS_DISK, "muF1", "MultiUser FFS disk" },
	{ ID_MU_OFS_INTL_DISK, "muF2", "MultiUser OFS INTL disk" },
	{ ID_MU_FFS_INTL_DISK, "muF3", "MultiUser FFS INTL disk" },
	{ ID_MU_OFS_DC_DISK, "muF4", "MultiUser OFS DC disk" },
	{ ID_MU_FFS_DC_DISK, "muF5", "MultiUser FFS DC disk" },
	{ 0, NULL, NULL }
};


	/*
	 *		Lilo Data
	 */

struct BootBlock BootBlock;

struct MapHeader MapHeader;
struct MapEntry *MapEntries = NULL;
u_long *MapVector;
int MapNumBlocks;

char *LoaderData;
int LoaderSize;
u_long *LoaderVector;
int LoaderNumBlocks;
u_long DosTypeID;


dev_t BootDevice;
struct stat info;
DIR *dir;
struct dirent *dirent;


	/*
	 *		Configuration File Parsing
	 */

#define arraysize(a) (sizeof(a)/sizeof(*(a)))
#define offsetof(t,e)	(unsigned long)(&(((t *)0)->e))


#define mygetc()						\
	({										\
		c = fgetc(CFile);				\
		if (c == '\n') ++LineNr;	\
		c;									\
	})
		
typedef enum {
	False = 0, True = 1
} BOOL;


/* ----------------------------------------------------------------- */
/*                      Configuration variables                      */

char *CnfDevice;
BOOL CnfAuto;
BOOL CnfAux;
int CnfTimeout;
char *CnfPassword;

typedef enum {
	INTEGER, STRING, BOOLEAN, BOOTOS
} VARTYPE;

typedef struct {
	const char *name;
	VARTYPE type;
	union {
		int i;
		char *cp;
		BOOL b;
	} dflt;
	void *var;
} CVAR;

CVAR ConfigVar[] = {
	{ "device",		STRING,	0,			&CnfDevice },
	{ "auto",		BOOLEAN,	False,	&CnfAuto },
	{ "aux",			BOOLEAN,	False,	&CnfAux },
	{ "timeout",	INTEGER,	20,		&CnfTimeout },
	{ "password",	STRING,	0,			&CnfPassword }
};


/* ----------------------------------------------------------------- */
/*                      Boot Record Definitions                      */

typedef struct _bootrec {
	struct _bootrec *next;
	char *image;			/* Kernel image */
	char *label;
	char *password;
	long type;
	char *args;				/* Arguments passed */
} BOOTREC;

BOOTREC	*BootRecs = NULL;


typedef struct {
	const char *name;
	VARTYPE type;
	unsigned	offset;
} BVAR;

BVAR BootrecVar[] = {
	{ "image",		STRING,	offsetof(BOOTREC, image) },
	{ "label",		STRING,	offsetof(BOOTREC, label) },
	{ "password",	STRING,	offsetof(BOOTREC, password) },
	{ "type",		BOOTOS,	offsetof(BOOTREC, type) },
	{ "args",		STRING,	offsetof(BOOTREC, args) }
};

#define BOOTRECREF(brp,type,off)	(*(type *)((char *)(brp) + (off)))

typedef struct {
	const char *name;
	long type;
} BOOTOSs;

BOOTOSs BootOSs[] = {
	{ "amiga",	BOS_AMIGA },
	{ "linux",	BOS_LINUX },
	{ "netbsd", BOS_NETBSD }
};


/* ----------------------------------------------------------------- */
/*                           Parser Types                            */

typedef enum {
	TOK_EOF, TOK_END, TOK_EQ, TOK_OPEN, TOK_CLOSE, TOK_INT, TOK_STRING
} TOKCLASS;

typedef struct {
	int i;
	char *cp;
	int b;
} TOKVAL;


/* ----------------------------------------------------------------- */
/*                        Global Variables                           */

FILE *CFile;
int LineNr = 1;
char c;					/* For mygetc(), if there is no local c */


	/*
	 *		Function Prototypes
	 */

static void Die(char *fmt, ...);
static void ReadBootBlock(char *name);
static void WriteBootBlock(char *name);
static int ReadFile(char *name, void **data);
static void WriteFile(char *name, void *data, int size);
static void ReadConfigFile(void);
static void CreateMapData(void);
static void WriteMapData(void);
static __inline u_long CalcChecksum(void);
static void GetGeometry(char *name, dev_t *device, u_long *start, u_long *blksize);
static u_long *CreateVector(char *name, int *numblocks);
static void PatchLoader(void);
static void CreateBootBlock();
static void Usage(void);
static void read_config_file(void);
static BOOL parse_vardef_or_bootdef(void);
static BOOL parse_vardef(TOKVAL *name);
static BOOL parse_bootdef(void);
static BOOL parse_bootrec(BOOTREC *br, TOKVAL *name);
static TOKCLASS gettok(TOKVAL *val);
static BOOTREC *new_bootrec(void);
static void add_bootrec(BOOTREC *br);
static char *safe_strdup(const char *p);
int main(int argc, char *argv[]);


	/*
	 *		Print an Error Message and Exit
	 */

static void Die(char *fmt, ...)
{
	va_list ap;

	fflush(stdout);
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	exit(1);
}


	/*
	 *		Read a Boot Block from a Device or File
	 */

static void ReadBootBlock(char *name)
{
	int fh;

	if ((fh = open(name, O_RDONLY)) == -1)
		Die("open %s: %s\n", name, strerror(errno));
	if (read(fh, &BootBlock, sizeof(BootBlock)) != sizeof(BootBlock))
		Die("read %s: %s\n", name, strerror(errno));
	close(fh);
}


	/*
	 *		Write a Boot Block to a Device
	 */

static void WriteBootBlock(char *name)
{
	int fh;

	if ((fh = open(name, O_RDWR)) == -1)
		Die("open %s: %s\n", name, strerror(errno));
	if (write(fh, &BootBlock, sizeof(BootBlock)) != sizeof(BootBlock))
		Die("write %s: %s\n", name, strerror(errno));
	close(fh);
}


	/*
	 *		Read a File
	 */

static int ReadFile(char *name, void **data)
{
	int fh, size = 0;

	if ((fh = open(name, O_RDONLY)) == -1)
		Die("open %s: %s\n", name, strerror(errno));
	if ((size = lseek(fh, 0, L_XTND)) == -1)
		Die("lseek %s: %s\n", name, strerror(errno));
	if (!(*data = malloc(size)))
		Die("No memory\n");
	if (lseek(fh, 0, L_SET) == -1)
		Die("lseek %s: %s\n", name, strerror(errno));
	if (read(fh, *data, size) != size)
		Die("read %s: %s\n", name, strerror(errno));
	close(fh);
	return(size);
}


	/*
	 *		Write a File
	 */

static void WriteFile(char *name, void *data, int size)
{
	int fh;

	if ((fh = open(name, O_CREAT|O_RDWR|O_TRUNC, S_IRUSR|S_IWUSR)) == -1)
		Die("open %s: %s\n", name, strerror(errno));
	if (write(fh, data, size) != size)
		Die("write %s: %s\n", name, strerror(errno));
	close(fh);
}


	/*
	 *		Read the Configuration File
	 */

static void ReadConfigFile(void)
{
	read_config_file();

	if (!CnfDevice)
		Die("No device specified\n");
	if (!BootRecs)
		Die("No boot records defined\n");

	Device = CnfDevice;	
}


	/*
	 *		Create the Map Data
	 */

static void CreateMapData(void)
{
	BOOTREC *br;
	struct MapEntry *entry;
	struct MapEntry **p;
	int size;
	int i = 0;
	u_long *vector;
	int numblocks;

	MapHeader.LiloID = LILO_ID;
	MapHeader.Auto = CnfAuto == True ? 1 : 0;
	MapHeader.Aux = CnfAux == True ? 1 : 0;
	MapHeader.TimeOut = CnfTimeout;
	if (CnfPassword) {
		if (strlen(CnfPassword) >= sizeof(MapHeader.MasterPwd))
			Die("Master Password is too long\n");
		strcpy(MapHeader.MasterPwd, CnfPassword);
	} else
		strcpy(MapHeader.MasterPwd, "");

	br = BootRecs;
	p = &MapEntries;
	do {
		if (br->type != BOS_AMIGA)
			if ((vector = CreateVector(br->image, &numblocks)))
				size = (numblocks+1)*sizeof(u_long);
			else {
				if (!i)
					Die("Default Operating System `%s' is not available\n", br->label);
				size = 0;
			}
		else {
			vector = NULL;
			size = 0;
		}
		entry = malloc(sizeof(struct MapEntry)+size);
		entry->Size = sizeof(struct MapEntry)+size;
		if (strlen(br->label) >= sizeof(entry->Label))
			Die("Label `%s' is too long\n", br->label);
		strcpy(entry->Label, br->label);
		if (br->password) {
			if (!i)
				Die("Default Operating System `%s' may not have a password\n", br->label);
			if (strlen(br->password) >= sizeof(entry->Password))
				Die("Password for `%s' is too long\n", br->label);
			strcpy(entry->Password, br->password);
		} else
			strcpy(entry->Password, "");
		if (vector)
			entry->Type = br->type;
		else
			entry->Type = -br->type;
		if (br->args) {
			if (strlen(br->args) >= sizeof(entry->CommandLine))
				Die("Command Line for `%s' is too long\n", br->label);
			strcpy(entry->CommandLine, br->args);
		} else
			strcpy(entry->CommandLine, "");
		if (vector) {
			memcpy(entry->KernelVector, vector, size);
			free(vector);
		}
		*p = entry;
		p = &entry->un.Next;
		i++;
	} while ((br = br->next));
	entry->un.Next = NULL;
}


	/*
	 *		Write the Map Data
	 */

static void WriteMapData(void)
{
	int fh;
	struct MapEntry *entry, *next;

	if ((fh = open(LILO_MAPFILE, O_CREAT|O_RDWR|O_TRUNC, S_IRUSR|S_IWUSR)) == -1)
		Die("open " LILO_MAPFILE ": %s\n", strerror(errno));
	if (write(fh, &MapHeader, sizeof(struct MapHeader)) != sizeof(struct MapHeader))
		Die("write " LILO_MAPFILE ": %s\n", strerror(errno));
	entry = MapEntries;
	puts("Boot labels:");
	do {
		next = entry->un.Next;
		entry->un.NotLast = next ? 1 : 0;
		printf("   %-32.32s (", entry->Label);
		if (entry->Type > 0)
			printf("%lu bytes)\n", entry->KernelVector[0]);
		else if (entry->Type == BOS_AMIGA)
			puts("ROM image)");
		else
			puts("not available)");
		if (write(fh, entry, entry->Size) != entry->Size)
			Die("write " LILO_MAPFILE ": %s\n", strerror(errno));
	} while ((entry = next));
	close(fh);
}


	/*
	 *		Calculate the Boot Block Checksum
	 */

static __inline u_long CalcChecksum(void)
{
	register u_long _res __asm("d0");
	register struct BootBlock *a0 __asm("a0") = &BootBlock;

	__asm __volatile ("moveq	#0,d0;"
							"moveq	#0,d1;"
							"not.b	d1;"
					"1:	 add.l	(a0)+,d0;"
							"jcc		2f;"
							"addq.l	#1,d0;"
					"2:	 dbf		d1,1b;"
							"not.l	d0;"
							: "=r" (_res)
							: "r" (a0)
							: "a0","a1","d0","d1","memory");
	return(_res);
}


	/*
	 *		Get the Geometry for the Partition a file belongs to
	 */

static void GetGeometry(char *name, dev_t *device, u_long *start, u_long *blksize)
{
	struct hd_geometry geo;
	char devname[PATH_MAX+1];
	int fh, found = 0;

	if (stat(name, &info) == -1)
		Die("stat %s: %s\n", name, strerror(errno));
	if (!S_ISREG(info.st_mode))
		Die("%s is not a regular file\n", name);
	*device = info.st_dev;
	*blksize = info.st_blksize;

	if (!(dir = opendir("/dev")))
		Die("opendir /dev: %s\n", strerror(errno));
	while (!found && (dirent = readdir(dir))) {
		sprintf(devname, "/dev/%s", dirent->d_name);
		if ((stat(devname, &info) != -1) && S_ISBLK(info.st_mode) && (info.st_rdev == *device)) {
			if ((fh = open(devname, O_RDONLY)) == -1)
				Die("open %s: %s\n", devname, strerror(errno));
			if (ioctl(fh, HDIO_GETGEO, &geo) == -1)
				Die("ioctl %s: %s\n", devname, strerror(errno));
			found = 1;
			close(fh);
		}
	}
	closedir(dir);
	if (found)
		*start = geo.start;
	else
		Die("Couldn't find special device file for file `%s'\n");
}


	/*
	 *		Create a Vector for a File
	 */

static u_long *CreateVector(char *name, int *numblocks)
{
	int fh, size, i;
	u_long *vector = NULL;
	u_long offset, start, blksize;
	dev_t device;
	int relblksize;

	if (!name || ((fh = open(name, O_RDONLY)) == -1))
		return(NULL);
	if ((size = lseek(fh, 0, L_XTND)) == -1)
		Die("lseek %s: %s\n", name, strerror(errno));
	*numblocks = (size+AMIGASECTORSIZE-1)/AMIGASECTORSIZE;
	if (!(vector = malloc((*numblocks+1)*sizeof(u_long))))
		Die("No memory\n");
	vector[0] = size;

	GetGeometry(name, &device, &start, &blksize);

	if (MAJOR(device) != MAJOR(BootDevice))
		Die("File `%s' must reside on physical device `%s'\n", name, Device);
 	switch (MAJOR(device)) {
		case FLOPPY_MAJOR:
			if (MINOR(device) != MINOR(BootDevice))
				Die("File `%s' must reside on physical floppy device `%s'\n", name, Device);
			break;

		case HD_MAJOR:
			if ((MINOR(device) & 0xc0) != (MINOR(BootDevice) & 0xc0))
				Die("File `%s' must reside on physical IDE device `%s'\n", name, Device);
			break;

		case SCSI_DISK_MAJOR:
			if ((MINOR(device) & 0xf0) != (MINOR(BootDevice) & 0xf0))
				Die("File `%s' must reside on physical SCSI device `%s'\n", name, Device);
			break;

		default:
			Die("Major device %d not supported (yet)\n", MAJOR(device));
			break;
	}
	relblksize = blksize/AMIGASECTORSIZE;

	for (i = 0; i < *numblocks; i++) {
		if (i % relblksize)
			offset = vector[i]+AMIGASECTORSIZE;
		else {
			offset = i/relblksize;
			if (ioctl(fh, FIBMAP, &offset) == -1)
				Die("ioctl FIBMAP %s: %s\n", name, strerror(errno));
			offset = start*AMIGASECTORSIZE+offset*blksize;
		}
		vector[i+1] = offset;
	}
	close(fh);
	return(vector);
}


	/*
	 *		Patch the Loader with the Map Vector
	 */

static void PatchLoader(void)
{
	int i;
	u_long pattern[] = {
		LILO_ID, LILO_MAPVECTORID
	};
	u_long *data = NULL;
	u_long maxsize;

	for (i = 0; i < LoaderSize-sizeof(pattern)+1; i++)
		if (!strncmp(&LoaderData[i], (char *)pattern, sizeof(pattern))) {
			data = (u_long *)&LoaderData[i];
			break;
		}
	if (!data)
		Die("Couldn't find magic key in loader template\n");
	maxsize = data[2];
	if (MapNumBlocks > maxsize-1)
		Die("Map vector is too large for loader (%d entries to much)\n", MapNumBlocks-maxsize-1);
	if (Verbose)
		printf("%d entries for map vector blocks, %lu entries free\n", MapNumBlocks, maxsize-1-MapNumBlocks);
	memcpy(&data[3], MapVector, (MapNumBlocks+1)*sizeof(u_long));
}


	/*
	 *		Create the Boot Block
	 */

static void CreateBootBlock()
{
	u_long maxsize, size;
	int i;

	if (DosType) {
		for (i = 0; DosTypeEntries[i].ID; i++)
			if (!strcmp(DosType, DosTypeEntries[i].ASCII))
				break;
		if (DosTypeEntries[i].ID)
			DosTypeID = DosTypeEntries[i].ID;
		else {
			printf("Unknown DosType `%s'\n", DosType);
			printf("Valid DosTypes are:\n");
			for (i = 0; DosTypeEntries[i].ID; i++)
				printf(" %s", DosTypeEntries[i].ASCII);
			Die("\n");
		}
	}

	size = BootCodeEnd-BootCodeStart;
	if (size > sizeof(BootBlock.Data))
		Die("%d bytes too much boot code\n", size-sizeof(BootBlock.Data));
	if (Verbose)
		printf("%lu bytes boot code, %lu bytes free\n", size, sizeof(BootBlock.Data)-size);
	BootBlock.ID = DosTypeID;
	BootBlock.Checksum = 0;
	if (BootBlock.ID == ID_BOOT_DISK)
		BootBlock.DosBlock = 0;
	memcpy(BootBlock.Data, BootCodeStart, size);

	maxsize = (LoaderVectorEnd-LoaderVectorStart)/sizeof(u_long);
	if (LoaderNumBlocks > maxsize-1)
		Die("Loader vector is too large for boot block (%d entries to much)\n", LoaderNumBlocks-maxsize-1);
	if (Verbose)
		printf("%d entries for loader vector blocks, %lu entries free\n", LoaderNumBlocks, maxsize-1-LoaderNumBlocks);
	memcpy((void *)((u_long)BootBlock.Data+(u_long)LoaderVectorStart-(u_long)BootCodeStart),
			 LoaderVector, (LoaderNumBlocks+1)*sizeof(u_long));

	BootBlock.LiloID = LILO_ID;
	BootBlock.Checksum = CalcChecksum();
}


	/*
	 *		Print the Usage Template and Exit
	 */

static void Usage(void)
{
	Die("Usage: %s [-v] [-f] [-t dostype] [-s bootblock] [-r bootblock] [-d device]\n", Name);
}


/* ================================================================= */
/*                             The Parser                            */

/*
 * Grammar of the config file:
 * ---------------------------
 *
 * Tokens are ';', '=', '{', '}', '#' and strings, which are a
 * none-white sequence of characters that are no separate tokens. If
 * blanks or token characters should be included in strings, enclose
 * them in '"'s. '#' starts a comment, which goes to the end of the
 * line.
 *
 * S ::= vardef S
 *     | bootdef S
 *     | eps
 *
 * vardef ::= string '=' string ';'
 *            (* The first string must be a valid variable name, the
 *			   * second a value that matches the variable's type. *)
 *
 * bootdef ::= "boot" '=' '{' bootreclist '}'
 *
 * bootreclist ::= bootrec bootreclist
 *               | eps
 *
 * bootrec ::= string '=' string ';'
 *             (* The first string must be a valid boot record
 *              * variable name, the second a value that matches the
 *              * variable's type. *)
 *
 * Additional semantic requirements:
 *  - There must be at last one boot record.
 *  - In each boot record, 'label' must be specified.
 *  
 */


static void read_config_file(void)
{
	if (!(CFile = fopen(LILO_CONFIGFILE, "r")))
		Die("open " LILO_CONFIGFILE ": %s\n", strerror(errno));

	while (!feof(CFile) && !ferror(CFile))
		parse_vardef_or_bootdef();

	fclose(CFile);
}


#define parse_err(text,sync)														\
	do {																					\
		fprintf(stderr, "Parse error in config file (line %d): %s\n",	\
				  LineNr, (text));													\
		if (*(sync))																	\
			while (!strchr((sync), mygetc()) && c != EOF)					\
				;																			\
		return(False);																	\
	} while (0)


/* ----------------------------------------------------------------- */
/*                       Non-Terminal functions                      */

static BOOL parse_vardef_or_bootdef(void)
{
	TOKVAL name;
	TOKCLASS class;
	
	class = gettok(&name);
	if (class == TOK_EOF)
		return(True);
	else if (class != TOK_STRING)
		parse_err("Identifier expected", ";");
	if (gettok(NULL) != TOK_EQ)
		parse_err("'=' expected after identifier", !strcmp(name.cp, "boot") ? "}" : ";");
	
	if (!strcmp(name.cp, "boot"))
		return(parse_bootdef());
	else
		return(parse_vardef(&name));
}


static BOOL parse_vardef(TOKVAL *name)
{
	int i;
	CVAR *cv;
	TOKCLASS class;
	TOKVAL val;
	
	for (i = 0; i < arraysize(ConfigVar); ++i)
		if (!strcmp(ConfigVar[i].name, name->cp))
			break;
	
	if (i >= arraysize(ConfigVar))
		parse_err("Unknown variable", ";");
	cv = &ConfigVar[i];

	class = gettok(&val);
	switch (cv->type) {
		case INTEGER:
			if (class != TOK_INT)
				parse_err("Value is no integer", ";");
			*(int *)(cv->var) = val.i;
			break;

		case STRING:
			*(char **)(cv->var) = safe_strdup(val.cp);
			break;

		case BOOLEAN:
			if (val.b == -1)
				parse_err("Value is not boolean", ";");
			*(BOOL *)(cv->var) = val.b;
			break;

		case BOOTOS:
			parse_err("Wrong type", ";");
	}

	if (gettok(NULL) != TOK_END)
		parse_err("';' missing at end of definition", "");

	return(True);
}


static BOOL parse_bootdef(void)
{
	TOKCLASS class;
	TOKVAL name;
	BOOTREC *br;
	
	if (gettok(NULL) != TOK_OPEN)
		parse_err("'{' expected after `boot ='", "}");

	br = new_bootrec();
	
	for (;;) {
		class = gettok(&name);
		if (class == TOK_CLOSE)
			break;
		else if (class == TOK_STRING) {
			if (!parse_bootrec(br, &name))
				return(False);
		} else
			parse_err("Identifier or '}' expected in boot record", ";}");
	}

	if (!br->label)
		parse_err("No `label' defined in boot record", "");
	
	add_bootrec(br);
	return(True);
}


static BOOL parse_bootrec(BOOTREC *br, TOKVAL *name)
{
	int i;
	BVAR *bv;
	TOKCLASS class;
	TOKVAL val;
	
	for (i = 0; i < arraysize(BootrecVar); ++i)
		if (!strcmp(BootrecVar[i].name, name->cp))
			break;
	
	if (i >= arraysize(BootrecVar))
		parse_err("Unknown variable", ";");
	bv = &BootrecVar[i];

	if (gettok(NULL) != TOK_EQ)
		parse_err("'=' expected after identifier", ";");
	
	class = gettok(&val);
	switch (bv->type) {
		case INTEGER:
			if (class != TOK_INT)
				parse_err("Value is no integer", ";");
			BOOTRECREF(br, int, bv->offset) = val.i;
			break;

		case STRING:
			BOOTRECREF(br, char *, bv->offset) = safe_strdup(val.cp);
			break;

		case BOOLEAN:
			if (val.b == -1)
				parse_err("Value is not boolean", ";");
			BOOTRECREF(br, BOOL, bv->offset) = val.b;
			break;

		case BOOTOS:
			for (i = 0; i < arraysize(BootOSs); i++)
				if (!strcasecmp(BootOSs[i].name, val.cp)) {
					BOOTRECREF(br, long, bv->offset) = BootOSs[i].type;
					break;
				}
			if (i == arraysize(BootOSs))
				parse_err("Invalid Operating System", ";");
			break;
	}

	if (gettok(NULL) != TOK_END)
		parse_err("';' missing at end of definition", "");

	return(True);
}



/* ----------------------------------------------------------------- */
/*                           Tokenizing                              */

static TOKCLASS gettok(TOKVAL *val)
{
	static char buf[1024];
	char *p, *endp, c;
	BOOL quote;
	int ival;

repeat:
	if (feof(CFile) || ferror(CFile))
		return(TOK_EOF);
	
	while (strchr(" \t\n", mygetc()))
		;
	
	if (c == EOF)
		return(TOK_EOF);
	else if (c == '#') {
		while (mygetc() != '\n' && c != EOF)
			;
		goto repeat;
	} else if (c == ';')
		return(TOK_END);
	else if (c == '=')
		return(TOK_EQ);
	else if (c == '{')
		return(TOK_OPEN);
	else if (c == '}')
		return(TOK_CLOSE);

	p = buf;
	quote = False;
	do {
		if (c == '"')
			quote = !quote;
		else
			*p++ = c;
	} while (p-buf < sizeof(buf)-1 && !(mygetc() == EOF || (!quote && strchr(" \t\n={};#", c))));
	*p = 0;
	if (c == EOF && quote)
		fprintf(stderr, "Parse error at end of config file: Unterminated string\n");
	if (c != EOF)
		ungetc(c, CFile);

	ival = strtol(buf, &endp, 0);
	if (val) {
		val->cp = buf;
		val->i = ival;
		val->b = (!strcmp(buf, "1") || !strcasecmp(buf, "true")  || !strcasecmp(buf, "t")) ? True :
					(!strcmp(buf, "0") || !strcasecmp(buf, "false") || !strcasecmp(buf, "f")) ? False : -1;
	}
	
	return((*endp == '\0') ? TOK_INT : TOK_STRING);
}



/* ----------------------------------------------------------------- */
/*                            Utilities...                           */

static BOOTREC *new_bootrec(void)
{
	BOOTREC *br;

	if (!(br = malloc(sizeof(BOOTREC))))
		Die("No memory\n");

	br->image = NULL;
	br->label = NULL;
	br->password = NULL;
	br->type = BOS_LINUX;
	br->args = NULL;

	return(br);
}


static void add_bootrec(BOOTREC *br)
{
	BOOTREC **brp;
	
	for (brp = &BootRecs; *brp; brp = &((*brp)->next))
		;
	*brp = br;
	br->next = NULL;
}


static char *safe_strdup(const char *p)
{
	char *q;
	
	if (!(q = strdup(p)))
		Die("No memory\n");

	return(q);
}


	/*
	 *		Main Routine
	 */

int main(int argc, char *argv[])
{
	int i;

	Name = argv[0];
	argc--;
	argv++;

	while (argc) {
		if (!strcmp(argv[0], "-v")) {
			puts(VERSION);
			Verbose = 1;
		} else if (!strcmp(argv[0], "-f"))
			Force = 1;
		else if (!strcmp(argv[0], "-t"))
			if (--argc && !DosType) {
				DosType = argv[1];
				argv++;
			} else
				Usage();
		else if (!strcmp(argv[0], "-s"))
			if (--argc && !SaveBootBlock) {
				SaveBootBlock = argv[1];
				Install = 0;
				argv++;
			} else
				Usage();
		else if (!strcmp(argv[0], "-r"))
			if (--argc && !RestoreBootBlock) {
				RestoreBootBlock = argv[1];
				Install = 0;
				argv++;
			} else
				Usage();
		else if (!strcmp(argv[0], "-d"))
			if (--argc && !Device) {
				Device = argv[1];
				argv++;
			} else
				Usage();
		else
			Usage();
		argc--;
		argv++;
	}

	if (Device && Install)
		Die("You can specify a device only with -s or -r\n");
	if (!Device && !Install)
		Die("You must specify a device if you use -s or -r\n");
	if (DosType && !Install)
		Die("You can't specify a dostype with -s or -r\n");

	if (Install)
		ReadConfigFile();

	if (stat(Device, &info) == -1)
		Die("stat %s: %s\n", Device, strerror(errno));
	if (!S_ISBLK(info.st_mode))
		Die("%s is not a block special device\n", Device);
	BootDevice = info.st_rdev;

	ReadBootBlock(Device);
	printf("%s: Boot block checksum = %s, ", Device, CalcChecksum() ? "Incorrect" : "OK");
	DosTypeID = BootBlock.ID;
	for (i = 0; DosTypeEntries[i].ID; i++)
		if (BootBlock.ID == DosTypeEntries[i].ID)
			break;
	if (DosTypeEntries[i].ID)
		printf("%s", DosTypeEntries[i].Description);
	else {
		printf("Type 0x%08lx", BootBlock.ID);
		if (!Force) {
			printf("\n");
			Die("Use option -f to force continue\n");
		}
		DosTypeID = ID_BOOT_DISK;
	}
	printf("%s\n", BootBlock.LiloID == LILO_ID ? " (created by Lilo)" : "");

	if (SaveBootBlock) {
		WriteFile(SaveBootBlock, &BootBlock, sizeof(BootBlock));
		printf("Boot block saved to file `%s'\n", SaveBootBlock);
	}
	if (RestoreBootBlock) {
		ReadBootBlock(RestoreBootBlock);
		printf("Boot block restored from file `%s'\n", RestoreBootBlock);
	}
	if (Install) {
		CreateMapData();
		WriteMapData();
		LoaderSize = ReadFile(LILO_LOADERTEMPLATE, (void **)&LoaderData);
		if (!(MapVector = CreateVector(LILO_MAPFILE, &MapNumBlocks)))
			Die("open `" LILO_MAPFILE "': %s\n", strerror(errno));
		if (Verbose)
			printf("Map vector has %lu bytes\n", MapVector[0]);
		PatchLoader();
		WriteFile(LILO_LOADERFILE, LoaderData, LoaderSize);
		if (!(LoaderVector = CreateVector(LILO_LOADERFILE, &LoaderNumBlocks)))
			Die("open `" LILO_LOADERFILE "': %s\n", strerror(errno));
		if (Verbose)
			printf("Loader vector has %lu bytes\n", LoaderVector[0]);
		CreateBootBlock();
	}
	if (RestoreBootBlock || Install) {
		WriteBootBlock(Device);
		printf("Boot block written to device `%s'\n", Device);
	}
	exit(0);
}
