/*
 *		Amiga Linux/68k Loader
 *
 *		 Copyright 1995 by Geert Uytterhoeven (Geert.Uytterhoeven@cs.kuleuven.ac.be)
 *
 *		-----------------------------------------------------------------------------
 *
 *		This program is based on files of the Linux/68k distribution:
 *
 *		 o bootstrap.c -- This program loads the Linux/68k kernel into an Amiga
 *								and launches it.
 *
 *			Copyright 1993, 1994 by Hamish Macdonald, Greg Harp
 *
 *			Modified 11-May-94 by Geert Uytterhoeven: A3640 MapROM check
 *			Modified 31-May-94 by Geert Uytterhoeven: Memory thrash problem solved
 *
 *		 o bootstrap.h -- This file is a part of the Amiga bootloader.
 *
 *			Copyright 1993, 1994 by Hamish Macdonald
 *
 *			Some minor additions by Michael Rausch 1-11-94
 *			Modified 11-May-94 by Geert Uytterhoeven: inline Supervisor() call
 *
 *
 *		The original idea for this program comes from the Linux distribution:
 *
 *			LILO  -  Generic Boot Loader for Linux ("LInux LOader")
 *						by Werner Almesberger
 *
 *		-----------------------------------------------------------------------------
 *
 *		This file is subject to the terms and conditions of the GNU General Public
 *		License.  See the file README.legal in the main directory of the Linux/68k
 *		distribution for more details.
 */


#ifndef __GNUC__
#error GNU CC is required to compile this program
#endif


#include <stddef.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>

#include <linux/a.out.h>
#include <linux/bootinfo.h>
#include <linux/amigatypes.h>
#include <linux/amigahw.h>

#include "loader.h"
#include "bootlib.h"
#include "lilo.h"


	/*
	 *		Lilo Alert
	 */

#define AN_LILO		AT_DeadEnd|0x4c000000			/* `L' */
#define AO_LiloMap	0x004d4150							/* `MAP': Corrupt Map File */


	/*
	 *		Console Echo Modes
	 */

#define ECHO_NONE			0
#define ECHO_NORMAL		1
#define ECHO_DOT			2


	/*
	 *		Program Entry Point (Called by the Boot Block Code)
	 */

asm("
	.text
_Init:
	movel	d0,_LiloTable;
	jmp	_Main
");


	/*
	 *		Function Prototypes
	 */

u_long Main(void);
static void ReadMapData(void);
static void FreeMapData(void);
static void OpenConsole(void);
static void CloseConsole();
static void ConsolePrintf(char *fmt, ...);
static void ConsolePuts(char *str);
static long ConsoleGetChar(void);
static void ConsolePutChar(char c);
static long ConsoleReadLine(u_long mode);
static void SetTimeOut(u_long seconds);
static void ClearTimeOut(void);
static void Sleep(u_long micros);
static struct MapEntry *GetBootOS(void);
static u_long BootAmigaOS(void);
static u_long BootLinux(u_long *kernelvector, char *commandline);
static u_long BootNetBSD(u_long *kernelvector, char *commandline);


	/*
	 *		Boot Library Functions
	 */

struct LiloTable *LiloTable;


	/*
	 *		Library Bases
	 */

struct ExecBase *SysBase;
struct ExpansionBase *ExpansionBase;
struct GfxBase *GfxBase;
struct IntuitionBase *IntuitionBase;
struct InputBase *InputBase;


	/*
	 *		Console Stuff
	 */

struct NewScreen NewScreen = {
	0, 0, 0, 0, 2, 0, 1, HIRES, CUSTOMSCREEN, NULL, NULL, NULL, NULL
};
struct NewWindow NewWindow = {
	0, 0, 0, 0, 2, 3, NULL, WFLG_SIMPLE_REFRESH|WFLG_ACTIVATE, NULL, NULL,
	"Amiga Linux/68k Loader", NULL, NULL, 0, 0, 0, 0, CUSTOMSCREEN
};
struct Screen *Screen;
struct Window *Window;
struct MsgPort *ConsolePort;
struct IOStdReq *ConsoleRequest;
struct MsgPort *TimerPort;
struct timerequest *TimerRequest;
struct MsgPort *InputPort;
struct IOStdReq *InputRequest;
u_long TimeOutSignalMask = NULL;
char FmtBuffer[256], InBuffer[CL_SIZE];
u_long ConsoleReady = FALSE;


	/*
	 *		Map Data
	 */

#define MAXVECTORSIZE	1024

struct FileVectorData {
	u_long ID[2];
	u_long MaxVectorSize;
	u_long Vector[MAXVECTORSIZE];
};

struct FileVectorData MapVectorData = {
	{	LILO_ID, LILO_MAPVECTORID	}, MAXVECTORSIZE
};

struct MapHeader MapHeader;
struct MapEntry *MapEntries = NULL;
u_short MasterMode = FALSE;


	/*
	 *		Linux/68k Bootstrap Stuff
	 */

/* temporary stack size */
#define TEMP_STACKSIZE	256

struct exec kexec;
char *memptr;
u_long start_mem;
u_long mem_size;
struct bootinfo bi;
u_long bi_size = sizeof(bi);
caddr_t CustomBase = (caddr_t)CUSTOM_PHYSADDR;


	/*
	 *		Main Loader Routine
	 *
	 *		This routine should return FALSE in case of an error (which will cause
	 *		a reboot), and TRUE if the user wants to boot AmigaOS instead of Linux.
	 */

u_long Main(void)
{
	u_long res = FALSE;
	struct MapEntry *bootentry = NULL;
	u_short keys;
	u_short interactive = FALSE;

	/* Open Libraries */

	SysBase = *(struct ExecBase **)4;
	if (!(ExpansionBase = (struct ExpansionBase *)OpenLibrary("expansion.library", 37)))
		Alert(AN_LILO|AG_OpenLib|AO_ExpansionLib);
	if (!(GfxBase = (struct GfxBase *)OpenLibrary("graphics.library", 37)))
		Alert(AN_LILO|AG_OpenLib|AO_GraphicsLib);

	/* Open Timer Device */

	if (!(TimerPort = CreateMsgPort()))
		Alert(AN_LILO|AG_NoSignal);
	if (!(TimerRequest = CreateIORequest(TimerPort, sizeof(struct timerequest))))
		Alert(AN_LILO|AG_NoMemory);
	if (OpenDevice("timer.device", UNIT_VBLANK, (struct IORequest *)TimerRequest, NULL))
		Alert(AN_LILO|AG_OpenDev|AO_TimerDev);

	/* Open Input Device */

	if (!(InputPort = CreateMsgPort()))
		Alert(AN_LILO|AG_NoSignal);
	if (!(InputRequest = CreateIORequest(InputPort, sizeof(struct IOStdReq))))
		Alert(AN_LILO|AG_NoMemory);
	if (OpenDevice("input.device", 0, (struct IORequest *)InputRequest, NULL))
		Alert(AN_LILO|AG_OpenDev|AO_Unknown);
	InputBase = (struct InputBase *)InputRequest->io_Device;

	/* Configuration */

	ReadMapData();
	keys = PeekQualifier();
	if (!MapHeader.Auto || keys)
		interactive = TRUE;

	if (interactive)
		OpenConsole();

	if (keys & IEQUALIFIER_CONTROL) {
		if (!MapHeader.MasterPwd[0])
			MasterMode = TRUE;
		else {
			ConsolePuts("Master password: ");
			if (((ConsoleReadLine(ECHO_DOT)) < 1) || strcmp(InBuffer, MapHeader.MasterPwd))
				ConsolePuts("\nIncorrect master password.\n");
			else {
				MasterMode = TRUE;
				ConsolePuts("\nEntering Master Mode.\n");
			}
		}
	}

	if (interactive)
		bootentry = GetBootOS();
	else
		bootentry = MapEntries;

	switch (bootentry->Type) {
		case BOS_AMIGA:
			ConsolePrintf("Booting AmigaOS (%s)...\n\n", bootentry->Label);
			res = BootAmigaOS();
			break;

		case BOS_LINUX:
			ConsolePrintf("Booting Linux/68k (%s)...\n\n", bootentry->Label);
			res = BootLinux(bootentry->KernelVector, bootentry->CommandLine);
			break;

		case BOS_NETBSD:
			ConsolePrintf("Booting NetBSD (%s)...\n\n", bootentry->Label);
			res = BootNetBSD(bootentry->KernelVector, bootentry->CommandLine);
			break;
	}
	if (!res) {
		if (bootentry->Type < 0)
			ConsolePuts("PANIC: Boot Operating System is not available\n");
		else
			ConsolePuts("PANIC: Couldn't boot desired Operating System\n");
		Sleep(2000000);
	}

	if (interactive)
		CloseConsole();

	FreeMapData();

	/* Close Input Device */

	CloseDevice((struct IORequest *)InputRequest);
	DeleteIORequest(InputRequest);
	DeleteMsgPort(InputPort);

	/* Close Timer Device */

	CloseDevice((struct IORequest *)TimerRequest);
	DeleteIORequest(TimerRequest);
	DeleteMsgPort(TimerPort);

	/* Close Libraries */

	CloseLibrary((struct Library*)GfxBase);
	CloseLibrary((struct Library *)ExpansionBase);
	return(res);
}


	/*
	 *		Read the Map Data
	 */

static void ReadMapData(void)
{
	u_long size;
	struct MapEntry *entry;

	Open(MapVectorData.Vector);
	Read(&MapHeader, sizeof(struct MapHeader));
	if (MapHeader.LiloID != LILO_ID)
		Alert(AN_LILO|AO_LiloMap);
	size = MapVectorData.Vector[0]-sizeof(struct MapHeader);
	if (!(MapEntries = (struct MapEntry *)AllocMem(size, MEMF_PUBLIC)))
		Alert(AN_LILO|AG_NoMemory);
	Read(MapEntries, size);
	Close();
	entry = MapEntries;
	do {
		if (entry->un.NotLast)
			entry->un.Next = (struct MapEntry *)((u_long)entry+entry->Size);
		else
			entry->un.Next = NULL;
	} while ((entry = entry->un.Next));
}


	/*
	 *		Free the Map Data
	 */

static void FreeMapData(void)
{
	if (MapEntries)
		FreeMem(MapEntries, MapVectorData.Vector[0]-sizeof(struct MapHeader));
}


	/*
	 *		Open the Console
	 */

static void OpenConsole(void)
{
	if (!(IntuitionBase = (struct IntuitionBase *)OpenLibrary("intuition.library", 37)))
		Alert(AN_LILO|AG_OpenLib|AO_Intuition);
	NewScreen.Width = GfxBase->NormalDisplayColumns;
	NewScreen.Height = GfxBase->NormalDisplayRows;
	if (!(Screen = OpenScreen(&NewScreen)))
		Alert(AN_LILO|AG_OpenRes|AO_Intuition);
	NewWindow.Screen = Screen;
	NewWindow.Width = NewScreen.Width;
	NewWindow.Height = NewScreen.Height;
	if (!(Window = OpenWindow(&NewWindow)))
		Alert(AN_LILO|AG_OpenRes|AO_Intuition);
	if (!(ConsolePort = CreateMsgPort()))
		Alert(AN_LILO|AG_NoSignal);
	if (!(ConsoleRequest = CreateIORequest(ConsolePort, sizeof(struct IOStdReq))))
		Alert(AN_LILO|AG_NoMemory);
	ConsoleRequest->io_Data = Window;
	if (OpenDevice("console.device", CONU_SNIPMAP, (struct IORequest *)ConsoleRequest, CONFLAG_DEFAULT))
		Alert(AN_LILO|AG_OpenDev|AO_ConsoleDev);
	ConsoleReady = TRUE;

	ConsolePuts("Amiga Linux/68k Loader Version 0.2\n");
	ConsolePuts(" Copyright 1995 by Geert Uytterhoeven\n\n\n");
}


	/*
	 *		Close the Console
	 */

static void CloseConsole()
{
	CloseDevice((struct IORequest *)ConsoleRequest);
	DeleteIORequest(ConsoleRequest);
	DeleteMsgPort(ConsolePort);
	CloseWindow(Window);
	CloseScreen(Screen);
	CloseLibrary((struct Library*)IntuitionBase);
}


	/*
	 *		Format and Print a String to the Console
	 */

asm("
.text
_PutCharInBuf:	cmp.l		#_FmtBuffer+255,a3
					jne		NotFull
					clr.b		(a3)
					rts
NotFull:			move.b	d0,(a3)+
					rts
");
extern void *PutCharInBuf();
	
static void ConsolePrintf(char *fmt, ...)
{
	va_list args;

	if (ConsoleReady) {
		va_start(args, fmt);
		RawDoFmt(fmt, args, (void (*)())&PutCharInBuf, FmtBuffer);
		va_end(args);
		ConsolePuts(FmtBuffer);
	}
}


	/*
	 *		Print a String to the Console
	 */

static void ConsolePuts(char *str)
{
	if (ConsoleReady) {
		ConsoleRequest->io_Command = CMD_WRITE;
		ConsoleRequest->io_Flags = IOF_QUICK;
		ConsoleRequest->io_Length = -1;
		ConsoleRequest->io_Data = str;
		DoIO((struct IORequest *)ConsoleRequest);
	}
}


	/*
	 *		Get a Character from the Console
	 */

static long ConsoleGetChar(void)
{
	long res = -1;
	u_long consolesig, sigs;
	char c;

	if (ConsoleReady) {
		ConsoleRequest->io_Command = CMD_READ;
		ConsoleRequest->io_Flags = NULL;
		ConsoleRequest->io_Length = 1;
		ConsoleRequest->io_Data = &c;
		SendIO((struct IORequest *)ConsoleRequest);
		consolesig = 1<<ConsolePort->mp_SigBit;
		for (;;) {
			sigs = Wait(consolesig|TimeOutSignalMask);
			if (sigs & consolesig) {
				res = c;
				break;
			}
			if (sigs & TimeOutSignalMask) {
				AbortIO((struct IORequest *)ConsoleRequest);
				break;
			}
		}
		WaitIO((struct IORequest *)ConsoleRequest);
	}
	return(res);
}


	/*
	 *		Put a Character to the Console
	 */

static void ConsolePutChar(char c)
{
	if (ConsoleReady) {
		ConsoleRequest->io_Command = CMD_WRITE;
		ConsoleRequest->io_Flags = IOF_QUICK;
		ConsoleRequest->io_Length = 1;
		ConsoleRequest->io_Data = &c;
		DoIO((struct IORequest *)ConsoleRequest);
	}
}


	/*
	 *		Read a Line from the Console
	 */

static long ConsoleReadLine(u_long mode)
{
	long c;
	long length = 0;
	u_long eol = FALSE;

	if (ConsoleReady) {
		while (!eol && ((c = ConsoleGetChar()) != -1))
			switch (c) {
				case '\r':
				case '\n':
					InBuffer[length] = '\0';
					if ((mode == ECHO_NORMAL) || (mode == ECHO_DOT))
						ConsolePutChar('\n');
					eol = TRUE;
					break;
				case '\b':
					if (length > 0) {
						length--;
						if ((mode == ECHO_NORMAL) || (mode == ECHO_DOT))
							ConsolePuts("\b \b");
					}
					break;
				default:
					if (length < sizeof(InBuffer)-1) {
						InBuffer[length++] = c;
						switch (mode) {
							case ECHO_NORMAL:
								ConsolePutChar(c);
								break;
							case ECHO_DOT:
								ConsolePutChar('');
								break;
						}
					}
					break;
			}
	}
	return(eol ? length : -1);
}


	/*
	 *		Set the Time Out Value
	 */

static void SetTimeOut(u_long seconds)
{
	TimerRequest->io_Command = TR_ADDREQUEST;
	TimerRequest->io_Flags = NULL;
	TimerRequest->tv_secs = seconds;
	TimerRequest->tv_micro = 0;
	SendIO((struct IORequest *)TimerRequest);
	TimeOutSignalMask = 1<<TimerPort->mp_SigBit;
}


	/*
	 *		Clear Time Out
	 */

static void ClearTimeOut(void)
{
	AbortIO((struct IORequest *)TimerRequest);
	WaitIO((struct IORequest *)TimerRequest);
	TimeOutSignalMask = NULL;
}


	/*
	 *		Sleep for a Specified Period
	 */

static void Sleep(u_long micros)
{
	TimerRequest->io_Command = TR_ADDREQUEST;
	TimerRequest->io_Flags = IOF_QUICK;
	TimerRequest->tv_secs = micros/1000000;
	TimerRequest->tv_micro = micros%1000000;
	DoIO((struct IORequest *)TimerRequest);
}


	/*
	 *		Ask the User for a Boot Operating System
	 */

static struct MapEntry *GetBootOS(void)
{
	struct MapEntry *bootentry = MapEntries;
	struct MapEntry *entry;
	long type;

	if (!MasterMode && MapHeader.TimeOut)
		SetTimeOut(MapHeader.TimeOut);

	while (1) {
		ConsolePuts("\n\nBoot image: ");
		if ((ConsoleReadLine(ECHO_NORMAL)) < 1)
			break;
		if (!strcmp(InBuffer, "help") || !strcmp(InBuffer, "?")) {
			ConsolePuts("\nList of boot images:\n\n");
			for (entry = MapEntries; entry; entry = entry->un.Next) {
				type = entry->Type;
				if (type < 0)
					type = -type;
				if (entry == MapEntries)
					ConsolePuts("[1m");
				ConsolePrintf("   %-31.31s (%s%s%s)\n", entry->Label,
								  type == BOS_AMIGA ? "AmigaOS" :
								  type == BOS_LINUX ? "Linux/68k" :
								  type == BOS_NETBSD ? "NetBSD-Amiga" : "Unknown",
								  entry->Password[0] ? ", Restricted" : "",
								  entry->Type < 0 ? ", N/A" : "");
				if (entry == MapEntries)
					ConsolePuts("[0m");
			}
		} else {
			for (entry = MapEntries; entry; entry = entry->un.Next)
				if (!strcmp(InBuffer, entry->Label))
					break;
			if (!entry) {
				ConsolePrintf("\nNonexistent boot image `%s'.\n", InBuffer);
				continue;
			} else if (entry->Type < 0) {
				ConsolePrintf("\nBoot image `%s' is not available.\n", InBuffer);
				continue;
			} else if (!MasterMode && entry->Password[0]) {
				ConsolePuts("\n\nPassword: ");
				if (((ConsoleReadLine(ECHO_DOT)) < 1) || strcmp(InBuffer, entry->Password)) {
					ConsolePuts("\nIncorrect password.\n");
					continue;
				}
			}
			bootentry = entry;
			break;
		}
	}

	if (MasterMode) {
		ConsolePrintf("\n\nOld command line: %s\n\nNew command line: ", bootentry->CommandLine);
		ConsoleReadLine(ECHO_NORMAL);
		strncpy(bootentry->CommandLine, InBuffer, CL_SIZE);
		bootentry->CommandLine[CL_SIZE-1] = '\0';
	}

	if (!MasterMode && MapHeader.TimeOut)
		ClearTimeOut();

	return(bootentry);
}


	/*
	 *		This assembler code is copied to chip ram, and then executed.
	 *		It copies the kernel to it's final resting place.
	 */

asm("
.text
.globl _copyall, _copyallend
_copyall:
										| /* put variables in registers because they may */
		lea		_kexec,a3		| /* be overwritten by kernel copy!! - G.U. */
		movel		_memptr,a4
		movel		_start_mem,a5
		movel		_mem_size,d0
		movel		_bi_size,d5
		movel		a3@(4),d2		| kexec.a_text
		movel		a3@(8),d3		| kexec.a_data
		movel		a3@(12),d4		| kexec.a_bss
	
										| /* copy kernel text and data */
		movel		a4,a0				| src = (u_long *)memptr;
		movel		a0,a2				| limit = (u_long *)(memptr+kexec.a_text+kexec.a_data);
		addl		d2,a2
		addl		d3,a2
		movel		a5,a1				| dest = (u_long *)start_mem;
1:		cmpl		a0,a2
		jeq		2f					| while (src < limit)
		moveb		a0@+,a1@+		|	*dest++ = *src++;
		jra		1b
2:
										| /* clear kernel bss */
		movel		a1,a0				| dest = (u_long *)(start_mem+kexec.a_text+kexec.a_data);
		movel		a1,a2				| limit = dest+kexec.a_bss / sizeof(u_long);
		addl		d4,a2
1:		cmpl		a0,a2
		jeq		2f					| while (dest < limit)
		clrb		a0@+				|	*dest++ = 0;
		jra		1b
2:

		movel		a4,a1				| src = (u long *)(memptr + kexec.a_text + kexec.a_data);
		addl		d2,a1
		addl		d3,a1				| dest = end of bss (already in a0)
		movel		d5,d7				| count = sizeof bi
		subql		#1,d7
1:		moveb		a1@+,a0@+		| while (--count > -1)
		dbra		d7,1b				|	*dest++ = *src++

										| /* jump to start of kernel */
		movel		a5,a0				| jump_to (START_MEM);
		jsr		a0@
_copyallend:
");


	/*
	 *		Test for a MapROMmed A3640 Board
	 */

asm("
.text
.globl _maprommed
_maprommed:
		oriw		#0x0700,sr
		moveml	#0x3f20,sp@-
										| /* Save cache settings */
		.long		0x4e7a1002		| movec cacr,d1 */
										| /* Save MMU settings */
		.long		0x4e7a2003		| movec tc,d2
		.long		0x4e7a3004		| movec itt0,d3
		.long		0x4e7a4005		| movec itt1,d4
		.long		0x4e7a5006		| movec dtt0,d5
		.long		0x4e7a6007		| movec dtt1,d6
		moveq		#0,d0
		movel		d0,a2
										| /* Disable caches */
		.long		0x4e7b0002		| movec d0,cacr
										| /* Disable MMU */
		.long		0x4e7b0003		| movec d0,tc
		.long		0x4e7b0004		| movec d0,itt0
		.long		0x4e7b0005		| movec d0,itt1
		.long		0x4e7b0006		| movec d0,dtt0
		.long		0x4e7b0007		| movec d0,dtt1
		lea		0x07f80000,a0
		lea		0x00f80000,a1
		movel		a0@,d7
		cmpl		a1@,d7
		jne		1f
		movel		d7,d0
		notl		d0
		movel		d0,a0@
		nop							| /* Thanks to Jrg Mayer! */
		cmpl		a1@,d0
		jne		1f
		moveq		#-1,d0			| /* MapROMmed A3640 present */
		movel		d0,a2
1:		movel		d7,a0@
										| /* Restore MMU settings */
		.long		0x4e7b2003		| movec d2,tc
		.long		0x4e7b3004		| movec d3,itt0
		.long		0x4e7b4005		| movec d4,itt1
		.long		0x4e7b5006		| movec d5,dtt0
		.long		0x4e7b6007		| movec d6,dtt1
										| /* Restore cache settings */
		.long		0x4e7b1002		| movec d1,cacr
		movel		a2,d0
		moveml	sp@+,#0x04fc
		rte
");

extern u_long maprommed();
extern char copyall, copyallend;


	/*
	 *		Boot The Standard Amiga Operating System
	 */

static u_long BootAmigaOS(void)
{
	return(TRUE);
}


	/*
	 *		Boot The Linux/68k Operating System
	 */

static u_long BootLinux(u_long *kernelvector, char *commandline)
{
	int debugflag = 1, i;
	long fast_total = 0;				/* total Fast RAM in system */
	struct MemHeader *mnp;
	struct ConfigDev *cdp = NULL;
	u_long memreq;
	void (*startfunc)(void);
	long startcodesize;
	u_long *stack;
	u_char chipset;

	/* print the greet message */
	ConsolePuts("\n\nLinux/68k Amiga Bootstrap version 1.10\n");
	ConsolePuts("Copyright 1993,1994 by Hamish Macdonald and Greg Harp\n\n");

	/* machine is Amiga */
	bi.machtype = MACH_AMIGA;


	/* Memory & AutoConfig based on 'unix_boot.c' by C= */

	/* find all of the autoconfig boards in the system */
	cdp = (struct ConfigDev *)FindConfigDev(cdp, -1, -1);
	for (i = 0; (i < NUM_AUTO) && cdp; i++) {
		/* copy the contents of each structure into our boot info */
		memcpy(&bi.bi_amiga.autocon[i], cdp, sizeof(struct ConfigDev));

		/* count this device */
		bi.bi_amiga.num_autocon++;

		/* get next device */
		cdp = (struct ConfigDev *)FindConfigDev(cdp, -1, -1);
	}

	/* find out the memory in the system */
	for (mnp = (struct MemHeader *)SysBase->MemList.lh_Head;
		  (bi.num_memory < NUM_MEMINFO) && mnp->mh_Node.ln_Succ;
		  mnp = (struct MemHeader *)mnp->mh_Node.ln_Succ) {
		struct MemHeader mh;

		/* copy the information */
		mh = *mnp;

		/* if we suspect that Kickstart is shadowed in an A3000,
			modify the entry to show 512K more at the top of RAM
			Check first for a MapROMmed A3640 board: overwriting the
			Kickstart image causes an infinite lock-up on reboot! */
		if (mh.mh_Upper == (void *)0x07f80000)
			if ((SysBase->AttnFlags & AFF_68040) && Supervisor(maprommed))
				ConsolePuts("A3640 MapROM detected.\n");
			else {
				mh.mh_Upper = (void *)0x08000000;
				ConsolePuts("A3000 shadowed Kickstart detected.\n");
			}

		/* if we suspect that Kickstart is zkicked,
			modify the entry to show 512K more at the botton of RAM */
		if (mh.mh_Lower == (void *)0x00280020) {
			mh.mh_Lower =  (void *)0x00200000;
			ConsolePuts("ZKick detected.\n");
		}

		/*
		 * If this machine has "LOCAL" memory between 0x07000000
		 * and 0x080000000, then we'll call it an A3000.
		 */
		if (mh.mh_Lower >= (void *)0x07000000 &&
			 mh.mh_Lower < (void *)0x08000000 &&
			 (mh.mh_Attributes & MEMF_LOCAL))
			bi.bi_amiga.model = AMI_3000;

		/* mask the memory limit values */
		mh.mh_Upper = (void *)((u_long)mh.mh_Upper & 0xfffff000);
		mh.mh_Lower = (void *)((u_long)mh.mh_Lower & 0xfffff000);

		/* if fast memory */
		if (mh.mh_Attributes & MEMF_FAST) {
			unsigned long size;

			/* record the start */
			bi.memory[bi.num_memory].addr = (u_long)mh.mh_Lower;

			/* set the size value to the size of this block */
			size = (u_long)mh.mh_Upper - (u_long)mh.mh_Lower;

			/* mask off to a 1M increment */
			size &= 0xfff00000;

			fast_total += size;

			if (size > 0)
				/* count this block */
				bi.memory[bi.num_memory++].size  = size;

		} else if (mh.mh_Attributes & MEMF_CHIP)
			/* if CHIP memory, record the size */
			bi.bi_amiga.chip_size = (u_long)mh.mh_Upper;
	}

	/* get info from ExecBase */
	bi.bi_amiga.vblank = SysBase->VBlankFrequency;
	bi.bi_amiga.psfreq = SysBase->PowerSupplyFrequency;
	bi.bi_amiga.eclock = SysBase->ex_EClockFrequency;

	/* Enable AGA (if possible) and determine chipset */
	if (GfxBase->Version >= 39)
		chipset = SetChipRev(SETCHIPREV_BEST);
	else
		chipset = GfxBase->ChipRevBits0;

	bi.bi_amiga.chipset = CS_STONEAGE;
	if ((chipset & GFXG_AGA) == GFXG_AGA) {
		bi.bi_amiga.chipset = CS_AGA;
		/*
		 *  we considered this machine to be an A3000 because of its
		 *  local memory just beneath $8000000; now if it has AGA, it
		 *  must be an A4000
		 *  except the case no RAM is installed on the motherboard but
		 *  on an additional card like FastLane Z3 or on the processor
		 *  board itself. Gotta check this out.
		 */
		bi.bi_amiga.model = (bi.bi_amiga.model == AMI_3000) ? AMI_4000 : AMI_1200;
	} else if ((chipset & GFXG_ECS) == GFXG_ECS)
		bi.bi_amiga.chipset = CS_ECS;
	else if ((chipset & GFXG_OCS) == GFXG_OCS)
		bi.bi_amiga.chipset = CS_OCS;

	/* Display amiga model */
	switch (bi.bi_amiga.model) {
		case AMI_UNKNOWN:
			break;
		case AMI_500:
			ConsolePuts("Amiga 500 ");
			break;
		case AMI_2000:
			ConsolePuts("Amiga 2000 ");
			break;
		case AMI_3000:
			ConsolePuts("Amiga 3000 ");
			break;
		case AMI_4000:
			ConsolePuts("Amiga 4000 ");
			break;
		case AMI_1200:								/* this implies an upgraded model   */
			ConsolePuts("Amiga 1200 ");		/* equipped with at least 68030 !!! */
			break;
	}

	/* display and set the CPU type */
	ConsolePuts("CPU: ");
	if (SysBase->AttnFlags & AFF_68040) {
		ConsolePuts("68040");
		bi.cputype = CPU_68040;
	} else if (SysBase->AttnFlags & AFF_68030) {
		ConsolePuts("68030");
		bi.cputype = CPU_68030;
	} else if (SysBase->AttnFlags & AFF_68020) {
		ConsolePuts("68020 (Do you have an MMU?)");
		bi.cputype = CPU_68020;
	} else {
		ConsolePuts("Insufficient for Linux.  Aborting...\n");
		ConsolePrintf("SysBase->AttnFlags = 0x%08lx\n", SysBase->AttnFlags);
		return(FALSE);
	}

	if ((SysBase->AttnFlags & (AFF_68040|AFF_FPU40)) == (AFF_68040|AFF_FPU40)) {
		ConsolePuts(" with internal FPU");
		bi.cputype |= FPU_68882;
	} else if (SysBase->AttnFlags & AFF_68882) {
		ConsolePuts(" with 68882 FPU");
		bi.cputype |= FPU_68882;
	} else if (SysBase->AttnFlags & AFF_68881) {
		ConsolePuts(" with 68881 FPU");
		bi.cputype |= FPU_68881;
	} else
		ConsolePuts(" without FPU");

	switch(bi.bi_amiga.chipset) {
		case CS_STONEAGE:
			ConsolePuts(", old or unknown chipset");
			break;
		case CS_OCS:
			ConsolePuts(", OCS");
			break;
		case CS_ECS:
			ConsolePuts(", ECS");
			break;
		case CS_AGA:
			ConsolePuts(", AGA chipset");
			break;
	}

	ConsolePuts("\n\n");

	/* copy command line options into the kernel command line */
	strncpy(bi.command_line, commandline, CL_SIZE);
	bi.command_line[CL_SIZE-1] = '\0';
	ConsolePrintf("Command line is '%s'\n", bi.command_line);

	/* display the clock statistics */
	ConsolePrintf("Vertical Blank Frequency: %ldHz\nPower Supply Frequency: %ldHz\n", bi.bi_amiga.vblank,
					  bi.bi_amiga.psfreq);
	ConsolePrintf("EClock Frequency: %ldKHz\n\n", (bi.bi_amiga.eclock+500)/1000);

	/* display autoconfig devices */
	if (bi.bi_amiga.num_autocon) {
		ConsolePrintf("Found %ld AutoConfig Device%s", bi.bi_amiga.num_autocon,
						  (bi.bi_amiga.num_autocon > 1) ? "s\n":"\n");
		for (i = 0; i < bi.bi_amiga.num_autocon; i++)
			ConsolePrintf("Device %ld: addr = %08lx\n", i, (u_long)bi.bi_amiga.autocon[i].cd_BoardAddr);
	} else
		ConsolePuts("No AutoConfig Devices Found\n");

	/* display memory */
	if (bi.num_memory) {
		ConsolePrintf("\n%ld Block%sof Memory Found\n", bi.num_memory, (bi.num_memory > 1)?"s ":" ");
		for (i = 0; i < bi.num_memory; i++)
			ConsolePrintf("Block %ld: %08lx to %08lx (%ldK)\n", i, bi.memory[i].addr,
							  bi.memory[i].addr+bi.memory[i].size, bi.memory[i].size >> 10);
	} else {
		ConsolePuts("No memory found?!  Aborting...\n");
		return(FALSE);
	}

	/* display chip memory size */
	ConsolePrintf("%ldK of CHIP memory\n", bi.bi_amiga.chip_size >> 10);

	start_mem = bi.memory[0].addr;
	mem_size = bi.memory[0].size;

	/* tell us where the kernel will go */
	ConsolePrintf("\nThe kernel will be located at %08lx\n", start_mem);

	/* verify that there is enough Chip RAM */
	if (bi.bi_amiga.chip_size < 512*1024) {
		ConsolePuts("\nNot enough Chip RAM in this system.  Aborting...\n");
		return(FALSE);
	}

	/* verify that there is enough Fast RAM */
	if (fast_total < 2*1024*1024) {
		ConsolePuts("\nNot enough Fast RAM in this system.  Aborting...\n");
		return(FALSE);
	}

	/* open kernel executable and read exec header */
	Open(kernelvector);
	Read((void *)&kexec, sizeof(kexec));
	switch (N_MAGIC(kexec)) {
		case ZMAGIC:
			ConsolePuts("ZMAGIC kernels are no longer supported\n");
			return(FALSE);
		case QMAGIC:
			/* the text size includes the exec header; remove this */
			kexec.a_text -= sizeof(kexec);
			break;
		default:
			ConsolePrintf("Wrong magic number 0x%08lx in kernel header\n", N_MAGIC(kexec));
			return(FALSE);
	}

	/* Load the kernel at one page after start of mem */
	start_mem += PAGE_SIZE;
	mem_size -= PAGE_SIZE;
	/* Align bss size to multiple of four */
	kexec.a_bss = (kexec.a_bss + 3) & ~3;

	/* no ram disk to boot from */
	bi.ramdisk_size = 0;
	bi.ramdisk_addr = (u_long)start_mem+mem_size;

	memreq = kexec.a_text+kexec.a_data+bi_size;
	if (!(memptr = (char *)AllocMem(memreq, MEMF_FAST|MEMF_CLEAR))) {
		ConsolePuts("Unable to allocate memory\n");
		return(FALSE);
	}

	/* data follows immediately after text */
	Read(memptr, kexec.a_text+kexec.a_data);
	Close();

	/* copy the boot_info struct to the end of the kernel image */
	memcpy((void *)(memptr+kexec.a_text+kexec.a_data), &bi, bi_size);

	/* allocate temporary chip ram stack */
	if (!(stack = (u_long *)AllocMem(TEMP_STACKSIZE, MEMF_CHIP|MEMF_CLEAR))) {
		ConsolePuts("Unable to allocate memory for stack\n");
		FreeMem((void *)memptr, memreq);
		return(FALSE);
	}

	/* allocate chip ram for copy of startup code */
	startcodesize = &copyallend - &copyall;
	if (!(startfunc = (void (*)(void))AllocMem(startcodesize, MEMF_CHIP|MEMF_CLEAR))) {
		ConsolePuts("Unable to allocate memory for code\n");
		FreeMem((void *)memptr, memreq);
		FreeMem((void *)stack, TEMP_STACKSIZE);
		return(FALSE);
	}

	/* copy startup code to CHIP RAM */
	memcpy(startfunc, &copyall, startcodesize);

	if (debugflag) {
		ConsolePrintf("\nKernel text at 0x%08lx, code size 0x%08lx\n", start_mem, kexec.a_text);
		ConsolePrintf("Kernel data at 0x%08lx, data size 0x%08lx\n", start_mem+kexec.a_text, kexec.a_data);
		ConsolePrintf("Kernel bss  at 0x%08lx, bss  size 0x%08lx\n", start_mem+kexec.a_text+kexec.a_data,
						  kexec.a_bss);
		ConsolePrintf("boot info   at 0x%08lx\n", start_mem+kexec.a_text+kexec.a_data+kexec.a_bss);
		ConsolePrintf("\nKernel entry is 0x%08lx\n", kexec.a_entry);
	}

	/* wait for things to settle down */
	Sleep(2000000);

	/* set graphics mode to a nice normal one */
	LoadView(NULL);

	Disable();

	/* Turn off all DMA */
	custom.dmacon = DMAF_ALL | DMAF_MASTER;

	/* turn off caches */
	CacheControl(NULL, ~NULL);

	/* Go into supervisor state */
	SuperState();

	/* setup stack */
	change_stack((char *)stack+TEMP_STACKSIZE);

	/* turn off any mmu translation */
	disable_mmu();

	/* execute the copy-and-go code (from CHIP RAM) */
	startfunc();

	/* NOTREACHED */
	return(TRUE);
}


	/*
	 *		Boot The NetBSD-Amiga Operating System
	 */

static u_long BootNetBSD(u_long *kernelvector, char *commandline)
{
	return(FALSE);
}
